from typing import Optional, List

from nonebot import get_driver
from pydantic import BaseSettings, validator, root_validator
from pydantic.fields import ModelField

from nonebot_plugin_pixivbot.enums import *
from nonebot_plugin_pixivbot.enums import DataSourceType
from nonebot_plugin_pixivbot.global_context import context


@context.register_singleton(**get_driver().config.dict())
class Config(BaseSettings):
    blacklist: set[str] = set()

    pixiv_refresh_token: str

    pixiv_data_source: DataSourceType = DataSourceType.sqlite
    pixiv_mongo_conn_url: Optional[str]
    pixiv_mongo_database_name: str
    pixiv_sql_conn_url: str = "sqlite+aiosqlite:///pixiv_bot.db"

    @root_validator(pre=True)
    def validate_data_source(cls, values):
        if "pixiv_data_source" not in values:
            if "pixiv_mongo_conn_url" in values:
                pixiv_data_source = DataSourceType.mongo
            else:
                pixiv_data_source = DataSourceType.sqlite
            values["pixiv_data_source"] = pixiv_data_source
        return values

    pixiv_proxy: Optional[str]
    pixiv_query_timeout: float = 60.0
    pixiv_loading_prompt_delayed_time: float = 5.0
    pixiv_simultaneous_query: int = 8

    pixiv_download_cache_expires_in = 3600 * 24 * 7
    pixiv_illust_detail_cache_expires_in = 3600 * 24 * 7
    pixiv_user_detail_cache_expires_in = 3600 * 24 * 7
    pixiv_illust_ranking_cache_expires_in = 3600 * 6
    pixiv_search_illust_cache_expires_in = 3600 * 24
    pixiv_search_illust_cache_delete_in = 3600 * 24 * 30
    pixiv_search_user_cache_expires_in = 3600 * 24
    pixiv_search_user_cache_delete_in = 3600 * 24 * 30
    pixiv_user_illusts_cache_expires_in = 3600 * 24
    pixiv_user_illusts_cache_delete_in = 3600 * 24 * 30
    pixiv_user_bookmarks_cache_expires_in = 3600 * 24
    pixiv_user_bookmarks_cache_delete_in = 3600 * 24 * 30
    pixiv_related_illusts_cache_expires_in = 3600 * 24
    pixiv_other_cache_expires_in = 3600 * 6

    pixiv_block_tags: List[str] = []
    pixiv_block_action: BlockAction = BlockAction.no_image

    pixiv_download_quantity: DownloadQuantity = DownloadQuantity.original
    pixiv_download_custom_domain: Optional[str]

    pixiv_compression_enabled: bool = False
    pixiv_compression_max_size: Optional[int]
    pixiv_compression_quantity: Optional[float]

    # 不加allow_reuse跑pytest会报错
    @validator('pixiv_compression_max_size', 'pixiv_compression_quantity', allow_reuse=True)
    def compression_validator(cls, v, values, field: ModelField):
        if values['pixiv_compression_enabled'] and v is None:
            raise ValueError(
                f'pixiv_compression_enabled is True but {field.name} got None.')
        return v

    # pixiv_illust_query_permission = {
    #     "group": [491959457],
    #     "friend": "all"
    # }
    #
    # @validator('pixiv_illust_query_permission')
    # def query_permission_validator(cls, v, field: ModelField):
    #     if v is not None and not isinstance(v, dict):
    #         raise ValueError(f'{field} expected a dict, but got a {type(v)}.')
    #     if "group" in v:
    #         if isinstance(v["group"], list):
    #             for i, x in enumerate(v["group"]):
    #                 if not isinstance(x, int):
    #                     raise ValueError(f'{field}["group"][{i}] expected a int, but got a {type(x)}.')
    #         elif v["group"] != "all":
    #             raise ValueError(f'{field}["group"] expected "all" or a list, but got a {type(v["group"])}.')
    #     if "friend" in v:
    #         if isinstance(v["friend"], list):
    #             for i, x in enumerate(v["group"]):
    #                 if not isinstance(x, int):
    #                     raise ValueError(f'{field}["friend"][{i}] expected a int, but got a {type(x)}.')
    #         elif v["friend"] != "all":
    #             raise ValueError(f'{field}["friend"] expected "all" or a list, but got a {type(v["friend"])}.')

    pixiv_query_to_me_only = False
    pixiv_command_to_me_only = False

    pixiv_query_cooldown = 0
    pixiv_no_query_cooldown_users: List[str] = []
    pixiv_max_item_per_query = 10

    pixiv_tag_translation_enabled = True

    pixiv_more_enabled = True
    pixiv_query_expires_in = 10 * 60

    pixiv_illust_query_enabled = True

    pixiv_ranking_query_enabled = True
    pixiv_ranking_default_mode: RankingMode = RankingMode.day
    pixiv_ranking_default_range = [1, 3]
    pixiv_ranking_fetch_item = 150
    pixiv_ranking_max_item_per_query = 10

    @validator('pixiv_ranking_default_range', allow_reuse=True)
    def ranking_default_range_validator(cls, v, field: ModelField):
        if len(v) < 2 or v[0] > v[1]:
            raise ValueError(f'illegal {field.name} value: {v}')
        return v

    pixiv_random_illust_query_enabled = True
    pixiv_random_illust_method = RandomIllustMethod.bookmark_proportion
    pixiv_random_illust_min_bookmark = 0
    pixiv_random_illust_min_view = 0
    pixiv_random_illust_max_page = 20
    pixiv_random_illust_max_item = 500

    pixiv_random_recommended_illust_query_enabled = True
    pixiv_random_recommended_illust_method = RandomIllustMethod.uniform
    pixiv_random_recommended_illust_min_bookmark = 0
    pixiv_random_recommended_illust_min_view = 0
    pixiv_random_recommended_illust_max_page = 40
    pixiv_random_recommended_illust_max_item = 1000

    pixiv_random_related_illust_query_enabled = True
    pixiv_random_related_illust_method = RandomIllustMethod.bookmark_proportion
    pixiv_random_related_illust_min_bookmark = 0
    pixiv_random_related_illust_min_view = 0
    pixiv_random_related_illust_max_page = 4
    pixiv_random_related_illust_max_item = 100

    pixiv_random_user_illust_query_enabled = True
    pixiv_random_user_illust_method = RandomIllustMethod.timedelta_proportion
    pixiv_random_user_illust_min_bookmark = 0
    pixiv_random_user_illust_min_view = 0
    pixiv_random_user_illust_max_page = 2 ** 31
    pixiv_random_user_illust_max_item = 2 ** 31

    pixiv_random_bookmark_query_enabled = True
    pixiv_random_bookmark_user_id: Optional[int] = None
    pixiv_random_bookmark_method = RandomIllustMethod.uniform
    pixiv_random_bookmark_min_bookmark = 0
    pixiv_random_bookmark_min_view = 0
    pixiv_random_bookmark_max_page = 2 ** 31
    pixiv_random_bookmark_max_item = 2 ** 31

    pixiv_random_following_illust_query_enabled = True
    pixiv_random_following_illust_method = RandomIllustMethod.timedelta_proportion
    pixiv_random_following_illust_min_bookmark = 0
    pixiv_random_following_illust_min_view = 0
    pixiv_random_following_illust_max_page = 2 ** 31
    pixiv_random_following_illust_max_item = 2 ** 31

    pixiv_watch_interval = 3600 * 2

    class Config:
        extra = "ignore"


__all__ = ("Config",)



from __future__ import annotations

from typing import Optional

from nonebot_plugin_pixivbot.model import Illust
from nonebot_plugin_pixivbot.utils.lazy_delegation import LazyDelegation

__all__ = ("LazyIllust",)


def _get_src():
    from .base import PixivRepo
    from nonebot_plugin_pixivbot import context
    return context.require(PixivRepo)


class LazyIllust:
    src = LazyDelegation(_get_src)

    def __init__(self, id: int, content: Optional[Illust] = None) -> None:
        self.id = id
        self.content = content

    async def get(self) -> Illust:
        if self.content is None:
            async for x in self.src.illust_detail(self.id):
                self.content = x
                break
        return self.content

    @property
    def loaded(self):
        return self.content is not None

    def __getattr__(self, attr):
        if self.content is None:
            return None
        else:
            return self.content.__getattribute__(attr)



from typing import TypeVar, Generic, Optional

from pydantic import root_validator
from pydantic.generics import GenericModel

UID = TypeVar("UID")
GID = TypeVar("GID")


class UserIdentifier(GenericModel, Generic[UID]):
    adapter: str
    user_id: UID

    def __init__(self, adapter: str, user_id: UID):
        super().__init__(adapter=adapter, user_id=user_id)

    def __str__(self):
        return f"{self.adapter}:{self.user_id}"

    class Config:
        frozen = True


class PostIdentifier(GenericModel, Generic[UID, GID]):
    adapter: str
    user_id: Optional[UID]
    group_id: Optional[GID]

    def __init__(self, adapter: str, user_id: Optional[UID] = None, group_id: Optional[GID] = None):
        super().__init__(adapter=adapter, user_id=user_id, group_id=group_id)

    def __str__(self):
        return f"{self.adapter}:{self.user_id}:{self.group_id}"

    @root_validator(allow_reuse=True)
    def validator(cls, values):
        if not values.get("user_id", None) and not values.get("group_id", None):
            raise ValueError("at least one of user_id and group_id should not be None")
        return values

    class Config:
        frozen = True



class QueryError(Exception):
    def __init__(self, message):
        self.message = message

    def __str__(self):
        return self.message


class RateLimitError(QueryError):
    def __init__(self):
        super(RateLimitError).__init__("Rate Limit")


class BadRequestError(Exception):
    def __init__(self, message=None):
        super().__init__()
        self.message = message

    def __str__(self):
        return self.message


__all__ = ("QueryError", "RateLimitError", "BadRequestError")



from typing import List, Union, Tuple

from nonebot import logger

from nonebot_plugin_pixivbot.config import Config
from nonebot_plugin_pixivbot.context import Inject
from nonebot_plugin_pixivbot.data.local_tag import LocalTagRepo
from nonebot_plugin_pixivbot.data.pixiv_repo import LazyIllust, PixivRepo
from nonebot_plugin_pixivbot.enums import RandomIllustMethod, RankingMode
from nonebot_plugin_pixivbot.global_context import context
from nonebot_plugin_pixivbot.model import Illust, User
from nonebot_plugin_pixivbot.service.roulette import roulette
from nonebot_plugin_pixivbot.utils.errors import BadRequestError, QueryError


@context.inject
@context.register_singleton()
class PixivService:
    conf = Inject(Config)
    repo = Inject(PixivRepo)
    local_tag_repo = Inject(LocalTagRepo)

    async def _choice_and_load(self, illusts: List[LazyIllust], random_method: RandomIllustMethod, count: int) \
            -> List[Illust]:
        if count <= 0:
            raise BadRequestError("不合法的请求数量")
        if count > self.conf.pixiv_max_item_per_query:
            raise BadRequestError("数量超过单次请求上限")
        if count > len(illusts):
            raise QueryError("别看了，没有的。")

        winners = roulette(illusts, random_method, count)
        logger.info(f"[pixiv_service] choice {[x.id for x in winners]}")
        return [await x.get() for x in winners]

    async def illust_ranking(self, mode: RankingMode, range: Tuple[int, int]) -> List[Illust]:
        # range 下标从1开始 闭区间
        i = 1
        li = []
        async for x in self.repo.illust_ranking(mode):
            if i > range[1]:
                break
            elif i >= range[0]:
                li.append(await x.get())
            i += 1
        return li

    async def illust_detail(self, illust: int) -> Illust:
        async for x in self.repo.illust_detail(illust):
            return x

    async def random_illust(self, word: str, *, count: int = 1) -> List[Illust]:
        if self.conf.pixiv_tag_translation_enabled:
            # 只有原word不是标签时获取翻译（例子：唐可可）
            tag = await self.local_tag_repo.find_by_name(word)
            if not tag:
                tag = await self.local_tag_repo.find_by_translated_name(word)
                if tag:
                    logger.info(f"[pixiv_service] found translation {word} -> {tag.name}")
                    word = tag.name

        illusts = [x async for x in self.repo.search_illust(word)]
        return await self._choice_and_load(illusts, self.conf.pixiv_random_illust_method, count)

    async def get_user(self, user: Union[str, int]) -> User:
        if isinstance(user, str):
            async for x in self.repo.search_user(user):
                logger.info(f"[pixiv_service] select user {x.name}({x.id})")
                return x
            raise QueryError("未找到用户")
        else:
            async for x in self.repo.user_detail(user):
                return x

    async def random_user_illust(self, user: Union[str, int], *, count: int = 1) -> Tuple[User, List[Illust]]:
        user = await self.get_user(user)
        illusts = [x async for x in self.repo.user_illusts(user.id)]
        illust = await self._choice_and_load(illusts, self.conf.pixiv_random_user_illust_method, count)
        return user, illust

    async def random_recommended_illust(self, *, count: int = 1) -> List[Illust]:
        illusts = [x async for x in self.repo.recommended_illusts()]
        return await self._choice_and_load(illusts, self.conf.pixiv_random_recommended_illust_method, count)

    async def random_bookmark(self, pixiv_user_id: int = 0, *, count: int = 1) -> List[Illust]:
        illusts = [x async for x in self.repo.user_bookmarks(pixiv_user_id)]
        return await self._choice_and_load(illusts, self.conf.pixiv_random_bookmark_method, count)

    async def random_related_illust(self, illust_id: int, *, count: int = 1) -> List[Illust]:
        if illust_id == 0:
            raise BadRequestError("你还没有发送过请求")

        illusts = [x async for x in self.repo.related_illusts(illust_id)]
        return await self._choice_and_load(illusts, self.conf.pixiv_random_related_illust_method, count)


__all__ = ("PixivService",)



from sqlalchemy import Column, Integer, select, String
from sqlalchemy.dialects.sqlite import insert

from nonebot_plugin_pixivbot import context
from nonebot_plugin_pixivbot.context import Inject
from nonebot_plugin_pixivbot.data.source.sql import SqlDataSource


@context.require(SqlDataSource).registry.mapped
class Seq:
    __tablename__ = "seq"

    key: str = Column(String, primary_key=True, nullable=False)
    value: int = Column(Integer, nullable=False, default=0)


@context.inject
@context.register_singleton()
class SqlSeqRepo:
    data_source: SqlDataSource = Inject(SqlDataSource)

    async def inc_and_get(self, key: str) -> int:
        session = self.data_source.session()
        stmt = (insert(Seq)
                .values(key=key, value=1)
                .on_conflict_do_update(index_elements=[Seq.key],
                                       set_={
                                           Seq.value: Seq.value + 1
                                       }))
        await session.execute(stmt)

        stmt = select(Seq).where(Seq.key == key)
        seq = (await session.execute(stmt)).scalar_one()
        return seq.value

    async def get_and_inc(self, key: str) -> int:
        return await self.inc_and_get(key) - 1

    async def get(self, key: str) -> int:
        session = self.data_source.session()
        stmt = select(Seq).where(Seq.key == key)
        seq = (await session.execute(stmt)).scalar_one_or_none()

        if seq is not None:
            return seq.value
        else:
            return 0



"""
为了让生命周期严格呈现以下模式：
on_startup -> on_bot_connect -> on_bot_disconnect -> on_shutdown
"""

import asyncio
from inspect import isawaitable

from nonebot import Bot, get_driver, logger

_on_startup_callback = []
_on_bot_connect_callback = []
_on_bot_disconnect_callback = []
_on_shutdown_callback = []

# 使用协程形式的Event是因为on_shutdown在on_bot_connected之前，如果用线程形式的Event会死锁
_startup = asyncio.Event()
_shutdown = asyncio.Event()

_mutex = asyncio.Lock()

_connected_bot = set()

_no_bot_connect = asyncio.Event()
_no_bot_connect.set()


# 注册回调不加锁是因为没有多线程场景

def on_startup(func, replay: bool = False):
    if replay and _startup.is_set():
        logger.debug("[lifecycler] replaying on_startup")
        x = func()
        if isawaitable(x):
            asyncio.create_task(x)
    else:
        _on_startup_callback.append(func)
    return func


def on_bot_connect(func, replay: bool = False):
    if replay:
        for bot in _connected_bot:
            logger.debug(f"[lifecycler] replaying on_bot_connect with {bot}")
            x = func(bot)
            if isawaitable(x):
                asyncio.create_task(x)
    _on_bot_connect_callback.append(func)
    return func


def on_bot_disconnect(func):
    _on_bot_disconnect_callback.append(func)
    return func


def on_shutdown(func):
    _on_shutdown_callback.append(func)
    return func


async def _fire_startup():
    await _mutex.acquire()
    try:
        logger.info("[lifecycler] on startup")
        cors = [f() for f in _on_startup_callback]
        cors = [x for x in cors if isawaitable(x)]
        if len(cors) > 0:
            await asyncio.gather(*cors)

        _startup.set()
    finally:
        _mutex.release()


async def _fire_bot_connect(bot: Bot):
    await _startup.wait()
    await _mutex.acquire()
    try:
        logger.info(f"[lifecycler] on bot {bot} connect")
        cors = [f(bot) for f in _on_bot_connect_callback]
        cors = [x for x in cors if isawaitable(x)]
        if len(cors) > 0:
            await asyncio.gather(*cors)

        _connected_bot.add(bot)
        _no_bot_connect.clear()
    finally:
        _mutex.release()


async def _fire_bot_disconnect(bot: Bot):
    await _startup.wait()
    await _mutex.acquire()
    try:
        logger.info(f"[lifecycler] on bot {bot} disconnect")
        cors = [f(bot) for f in _on_bot_disconnect_callback]
        cors = [x for x in cors if isawaitable(x)]
        if len(cors) > 0:
            await asyncio.gather(*cors)

        _connected_bot.remove(bot)
        if not _connected_bot:
            _no_bot_connect.set()
    finally:
        _mutex.release()


async def _fire_shutdown():
    await _no_bot_connect.wait()
    await _mutex.acquire()
    try:
        logger.info(f"[lifecycler] on shutdown")
        cors = [f() for f in _on_shutdown_callback]
        cors = [x for x in cors if isawaitable(x)]
        if len(cors) > 0:
            await asyncio.gather(*cors)
    finally:
        _mutex.release()


_driver = get_driver()
_driver.on_startup(_fire_startup)
_driver.on_bot_connect(_fire_bot_connect)
_driver.on_bot_disconnect(_fire_bot_disconnect)
_driver.on_shutdown(_fire_shutdown)
